import torch

__all__ = [
    'ARRAY_TYPE',
    'DTYPE_TYPE',
    'DEVICE_TYPE',
    'RNG_TYPE',
]

ARRAY_TYPE = torch.Tensor
DTYPE_TYPE = torch.dtype
DEVICE_TYPE = torch.device
RNG_TYPE = torch.Generator